@@ -21,6 +21,10 @@
 #include <signal.h>
 #include <sys/stat.h>
 #include <sys/wait.h>
+#include <sys/socket.h>
+#include <netinet/in.h>
+#include <arpa/inet.h>
+#include <netdb.h>
 #include <ctype.h>
 #include <pwd.h>
 #include <grp.h>
@@ -35,23 +39,30 @@
 #include "xdmcp.h"
 #include "verify.h"
 #include "display.h"
+#include "choose.h"
 #include "errorgui.h"
 
-static const gchar RCSid[]="$Id: gdm.c,v 1.49 2001/07/02 21:10:54 jirka Exp $";
+static const gchar RCSid[]="$Id: gdm.c,v 1.53 2001/07/06 08:53:56 jirka Exp $";
 
 
 /* Local functions */
 static void gdm_config_parse (void);
 static void gdm_local_servers_start (GdmDisplay *d);
 static void gdm_daemonify (void);
+static gboolean gdm_slave_socket_handler (GIOChannel *source,
+					  GIOCondition cond,
+					  gpointer data);
+static void gdm_safe_restart (void);
+static void gdm_restart_now (void);
 
 /* Global vars */
 GSList *displays;		/* List of displays managed */
 gint sessions = 0;		/* Number of remote sessions */
 sigset_t sysmask;		/* Inherited system signal mask */
-gchar *argdelim = " ";		/* argv argument delimiter */
 uid_t GdmUserId;		/* Userid under which gdm should run */
 gid_t GdmGroupId;		/* Groupid under which gdm should run */
+int fifofd = -1;		/* fifo fd for the slave communication fifo */
+guint fifo_source = 0;		/* source for the above */
 
 /* True if the server that was run was in actuallity not specified in the
  * config file.  That is if xdmcp was disabled and no local servers were
@@ -128,6 +139,8 @@ char **stored_argv = NULL;
 int stored_argc = 0;
 char *stored_path = NULL;
 
+static gboolean gdm_restart_mode = FALSE;
+
 
 /**
  * gdm_config_parse:
@@ -143,7 +156,7 @@ gdm_config_parse (void)
     struct passwd *pwent;
     struct group *grent;
     struct stat statbuf;
-    gchar *gtemp, *stemp;
+    gchar *bin;
     
     displays = NULL;
 
@@ -351,31 +364,26 @@ gdm_config_parse (void)
 
 
     /* Check that the greeter can be executed */
-    gtemp = g_strdup (GdmGreeter);
-    stemp = strchr (gtemp, ' ');
+    bin = gdm_first_word (GdmGreeter);
 
-    if (stemp) {
-	*stemp = '\0';
-
-	if (access (gtemp, X_OK))
-	    gdm_error ("gdm_config_parse: Greeter not found or can't be executed by the gdm user", gtemp);
+    if ( ! gdm_string_empty (bin) &&
+	access (bin, X_OK) != 0) {
+	    gdm_error (_("%s: Greeter not found or can't be executed by the gdm user"), "gdm_config_parse");
     }
 
-    g_free (gtemp);
+    g_free (bin);
 
 
     /* Check that chooser can be executed */
-    gtemp = g_strdup (GdmChooser);
-    stemp = strchr (gtemp, ' ');
+    bin = gdm_first_word (GdmChooser);
 
-    if (GdmIndirect && stemp) {
-	*stemp = '\0';
-
-	if (access (gtemp, X_OK))
-	    gdm_error ("gdm_config_parse: Chooser not found or it can't be executed by the gdm user", gtemp);
+    if (GdmIndirect &&
+	! gdm_string_empty (bin) &&
+	access (bin, X_OK) != 0) {
+	    gdm_error (_("%s: Chooser not found or it can't be executed by the gdm user"), "gdm_config_parse");
     }
     
-    g_free (gtemp);
+    g_free (bin);
 
 
     /* Enter paranoia mode */
@@ -436,9 +444,9 @@ gdm_daemonify (void)
     close (1);
     close (2);
 
-    open ("/dev/null", O_RDONLY);
-    dup2 (0, 1);
-    dup2 (0, 2);
+    open("/dev/null", O_RDONLY); /* open stdin - fd 0 */
+    open("/dev/null", O_RDWR); /* open stdout - fd 1 */
+    open("/dev/null", O_RDWR); /* open stderr - fd 2 */
 }
 
 
@@ -485,6 +493,7 @@ final_cleanup (void)
 	g_slist_free (list);
 
 	gdm_xdmcp_close ();
+	gdm_fifo_close ();
 
 	closelog();
 	unlink (GdmPidFile);
@@ -498,12 +507,10 @@ deal_with_x_crashes (GdmDisplay *d)
 
     if ( ! d->failsafe_xserver &&
 	 ! gdm_string_empty (GdmFailsafeXServer)) {
-	    char *bin = g_strdup (GdmFailsafeXServer);
-	    char *p = strchr (bin, argdelim[0]);
-	    if (p != NULL)
-		    *p = '\0';
+	    char *bin = gdm_first_word (GdmFailsafeXServer);
 	    /* Yay we have a failsafe */
-	    if (access (bin, X_OK) == 0) {
+	    if ( ! gdm_string_empty (bin) &&
+		access (bin, X_OK) == 0) {
 		    gdm_info (_("deal_with_x_crashes: Trying failsafe X "
 				"server %s"), GdmFailsafeXServer);
 		    g_free (bin);
@@ -525,8 +532,7 @@ deal_with_x_crashes (GdmDisplay *d)
 	    char **configurators;
 	    int i;
 
-	    configurators = g_strsplit (GdmXKeepsCrashingConfigurators,
-					argdelim, MAX_ARGS);	
+	    configurators = gdm_split (GdmXKeepsCrashingConfigurators);
 	    for (i = 0; configurators[i] != NULL; i++) {
 		    if (access (configurators[i], X_OK) == 0)
 			    break;
@@ -658,7 +664,7 @@ bin_executable (const char *command)
 	if (gdm_string_empty (command))
 		return FALSE;
 
-	argv = g_strsplit (command, argdelim, MAX_ARGS);	
+	argv = gdm_split (command);
 	if (argv != NULL &&
 	    argv[0] != NULL &&
 	    access (argv[0], X_OK) == 0) {
@@ -677,16 +683,21 @@ gdm_cleanup_children (void)
     gint exitstatus = 0, status;
     GdmDisplay *d = NULL;
     gchar **argv;
+    gboolean crashed;
 
     /* Pid and exit status of slave that died */
     pid = waitpid (-1, &exitstatus, WNOHANG);
 
-    if (WIFEXITED (exitstatus))
+    if (WIFEXITED (exitstatus)) {
 	    status = WEXITSTATUS (exitstatus);
-    else
+	    crashed = FALSE;
+    } else {
 	    status = DISPLAY_SUCCESS;
+	    crashed = TRUE;
+    }
 	
-    gdm_debug ("gdm_cleanup_children: child %d returned %d", pid, status);
+    gdm_debug ("gdm_cleanup_children: child %d returned %d%s", pid, status,
+	       crashed ? " (child crashed)" : "");
 
     if (pid < 1)
 	return;
@@ -697,6 +708,41 @@ gdm_cleanup_children (void)
     if (!d)
 	return;
 
+    if (crashed) {
+	    gdm_debug ("gdm_cleanup_children: Slave crashed, killing it's "
+		       "children");
+
+	    if (d->sesspid > 0) {
+		    if (kill (d->sesspid, SIGTERM) == 0)
+			    waitpid (d->sesspid, NULL, 0);
+		    d->sesspid = 0;
+	    }
+	    if (d->greetpid > 0) {
+		    if (kill (d->greetpid, SIGTERM) == 0)
+			    waitpid (d->greetpid, NULL, 0);
+		    d->greetpid = 0;
+	    }
+	    if (d->chooserpid > 0) {
+		    if (kill (d->chooserpid, SIGTERM) == 0)
+			    waitpid (d->chooserpid, NULL, 0);
+		    d->chooserpid = 0;
+	    }
+	    if (d->servpid > 0) {
+		    if (kill (d->servpid, SIGTERM) == 0)
+			    waitpid (d->servpid, NULL, 0);
+		    d->servpid = 0;
+	    }
+    }
+
+    /* null all these, they are not valid most definately */
+    d->servpid = 0;
+    d->sesspid = 0;
+    d->greetpid = 0;
+    d->chooserpid = 0;
+
+    /* definately not logged in now */
+    d->logged_in = FALSE;
+
     /* Declare the display dead */
     d->slavepid = 0;
     d->dispstat = DISPLAY_DEAD;
@@ -748,6 +794,9 @@ gdm_cleanup_children (void)
     case DISPLAY_ABORT:		/* Bury this display for good */
 	gdm_info (_("gdm_child_action: Aborting display %s"), d->name);
 
+	if (gdm_restart_mode)
+		gdm_safe_restart ();
+
 	gdm_display_unmanage (d);
 	break;
 	
@@ -756,7 +805,7 @@ gdm_cleanup_children (void)
 
 	final_cleanup ();
 
-	argv = g_strsplit (GdmReboot, argdelim, MAX_ARGS);	
+	argv = gdm_split (GdmReboot);
 	execv (argv[0], argv);
 
 	gdm_error (_("gdm_child_action: Reboot failed: %s"), strerror (errno));
@@ -767,7 +816,7 @@ gdm_cleanup_children (void)
 
 	final_cleanup ();
 
-	argv = g_strsplit (GdmHalt, argdelim, MAX_ARGS);	
+	argv = gdm_split (GdmHalt);
 	execv (argv[0], argv);
 
 	gdm_error (_("gdm_child_action: Halt failed: %s"), strerror (errno));
@@ -778,23 +827,20 @@ gdm_cleanup_children (void)
 
 	final_cleanup ();
 
-	argv = g_strsplit (GdmReboot, argdelim, MAX_ARGS);	
+	argv = gdm_split (GdmSuspend);
 	execv (argv[0], argv);
 
 	gdm_error (_("gdm_child_action: Suspend failed: %s"), strerror (errno));
 	break;
-	
 
     case DISPLAY_RESTARTGDM:
-	final_cleanup ();
-	if (stored_path != NULL)
-		putenv (stored_path);
-	execvp (stored_argv[0], stored_argv);
-	gdm_error (_("Failed to restart self"));
-	_exit (1);
+	gdm_restart_now ();
 	break;
 
     case DISPLAY_XFAILED:       /* X sucks */
+	if (gdm_restart_mode)
+		gdm_safe_restart ();
+
 	/* in remote case just drop to _REMANAGE */
 	if (d->type == TYPE_LOCAL) {
 		time_t now = time (NULL);
@@ -828,6 +874,9 @@ gdm_cleanup_children (void)
     case DISPLAY_REMANAGE:	/* Remanage display */
     default:
 	gdm_debug ("gdm_child_action: Slave process returned %d", status);
+
+	if (gdm_restart_mode)
+		gdm_safe_restart ();
 	
 	/* This is a local server so we start a new slave */
 	if (d->type == TYPE_LOCAL) {
@@ -841,18 +890,38 @@ gdm_cleanup_children (void)
 	break;
     }
 
+    if (gdm_restart_mode)
+	    gdm_safe_restart ();
+
     gdm_quit ();
 }
 
 static void
-term_cleanup (void)
+gdm_restart_now (void)
 {
-	gdm_debug ("term_cleanup: Got TERM/INT. Going down!");
-
+	gdm_info (_("Gdm restarting ..."));
 	final_cleanup ();
+	if (stored_path != NULL)
+		putenv (stored_path);
+	execvp (stored_argv[0], stored_argv);
+	gdm_error (_("Failed to restart self"));
+	_exit (1);
 }
 
+static void
+gdm_safe_restart (void)
+{
+	GSList *li;
 
+	for (li = displays; li != NULL; li = li->next) {
+		GdmDisplay *d = li->data;
+
+		if (d->logged_in)
+			return;
+	}
+
+	gdm_restart_now ();
+}
 
 static gboolean
 mainloop_sig_callback (gint8 sig, gpointer data)
@@ -866,17 +935,18 @@ mainloop_sig_callback (gint8 sig, gpointer data)
 
     case SIGINT:
     case SIGTERM:
-      term_cleanup ();
+      gdm_debug ("mainloop_sig_callback: Got TERM/INT. Going down!");
+      final_cleanup ();
       exit (EXIT_SUCCESS);
       break;
 
     case SIGHUP:
-      term_cleanup ();
-      if (stored_path != NULL)
-	      putenv (stored_path);
-      execvp (stored_argv[0], stored_argv);
-      gdm_error (_("Failed to restart self"));
-      _exit (1);
+      gdm_restart_now ();
+      break;
+
+    case SIGUSR1:
+      gdm_restart_mode = TRUE;
+      gdm_safe_restart ();
       break;
  
     default:
@@ -912,6 +982,42 @@ store_argv (int argc, char *argv[])
 	stored_argc = argc;
 }
 
+static void
+create_fifo (void)
+{
+	gchar *fifopath;
+	GIOChannel *fifochan;
+
+	fifopath = g_strconcat (GdmServAuthDir, "/.gdmfifo", NULL);
+
+	unlink (fifopath);
+
+	if (mkfifo (fifopath, 0660) < 0) {
+		gdm_fail (_("%s: Could not make FIFO"),
+			  "create_fifo");
+		return;
+	}
+
+	fifofd = open (fifopath, O_RDWR); /* Open with write to avoid EOF */
+
+	if (fifofd < 0) {
+		gdm_fail (_("%s: Could not open FIFO"),
+			  "create_fifo");
+		return;
+	}
+
+	chmod (fifopath, 0660);
+
+	g_free (fifopath);
+
+	fifochan = g_io_channel_unix_new (fifofd);
+	fifo_source = g_io_add_watch_full
+		(fifochan, G_PRIORITY_DEFAULT,
+		 G_IO_IN|G_IO_PRI|G_IO_ERR|G_IO_HUP|G_IO_NVAL,
+		 gdm_slave_socket_handler, NULL, NULL);
+	g_io_channel_unref (fifochan);
+}
+
 int 
 main (int argc, char *argv[])
 {
@@ -989,19 +1095,27 @@ main (int argc, char *argv[])
     g_signal_add (SIGTERM, mainloop_sig_callback, NULL);
     g_signal_add (SIGINT, mainloop_sig_callback, NULL);
     g_signal_add (SIGHUP, mainloop_sig_callback, NULL);
+    g_signal_add (SIGUSR1, mainloop_sig_callback, NULL);
     
     term.sa_handler = signal_notify;
     term.sa_flags = SA_RESTART;
     sigemptyset (&term.sa_mask);
 
     if (sigaction (SIGTERM, &term, NULL) < 0) 
-	gdm_fail (_("gdm_main: Error setting up TERM signal handler"));
+	gdm_fail (_("%s: Error setting up TERM signal handler"),
+		  "gdm_main");
 
     if (sigaction (SIGINT, &term, NULL) < 0) 
-	gdm_fail (_("gdm_main: Error setting up INT signal handler"));
+	gdm_fail (_("%s: Error setting up INT signal handler"),
+		  "gdm_main");
 
     if (sigaction (SIGHUP, &term, NULL) < 0) 
-	gdm_fail (_("gdm_main: Error setting up HUP signal handler"));
+	gdm_fail (_("%s: Error setting up HUP signal handler"),
+		  "gdm_main");
+
+    if (sigaction (SIGUSR1, &term, NULL) < 0) 
+	gdm_fail (_("%s: Error setting up USR1 signal handler"),
+		  "gdm_main");
 
     child.sa_handler = signal_notify;
     child.sa_flags = SA_RESTART|SA_NOCLDSTOP;
@@ -1009,13 +1123,14 @@ main (int argc, char *argv[])
     sigaddset (&child.sa_mask, SIGCHLD);
 
     if (sigaction (SIGCHLD, &child, NULL) < 0) 
-	gdm_fail (_("gdm_main: Error setting up CHLD signal handler"));
+	gdm_fail (_("%s: Error setting up CHLD signal handler"), "gdm_main");
 
     sigemptyset (&mask);
     sigaddset (&mask, SIGINT);
     sigaddset (&mask, SIGTERM);
     sigaddset (&mask, SIGCHLD);
     sigaddset (&mask, SIGHUP);
+    sigaddset (&mask, SIGUSR1);
     sigprocmask (SIG_UNBLOCK, &mask, &sysmask); /* Save system sigmask */
 
     gdm_debug ("gdm_main: Here we go...");
@@ -1024,6 +1139,8 @@ main (int argc, char *argv[])
     if (GdmXdmcp)
 	gdm_xdmcp_init();
 
+    create_fifo ();
+
     /* Start local X servers */
     g_slist_foreach (displays, (GFunc) gdm_local_servers_start, NULL);
 
@@ -1161,4 +1278,167 @@ gdm_quit (void)
   g_main_quit (main_loop);
 }
 
+static void
+handle_message (const char *msg)
+{
+	gdm_debug ("Handeling message: '%s'", msg);
+
+	if (strncmp (msg, GDM_SOP_CHOSEN " ",
+		     strlen (GDM_SOP_CHOSEN " ")) == 0) {
+		gdm_choose_data (msg);
+	} else if (strncmp (msg, GDM_SOP_XPID " ",
+		     strlen (GDM_SOP_XPID " ")) == 0) {
+		GdmDisplay *d;
+		long slave_pid, pid;
+
+		if (sscanf (msg, GDM_SOP_XPID " %ld %ld", &slave_pid, &pid)
+		    != 2)
+			return;
+
+		/* Find out who this slave belongs to */
+		d = gdm_display_lookup (slave_pid);
+
+		if (d != NULL) {
+			d->servpid = pid;
+			gdm_debug ("Got XPID == %ld", (long)pid);
+			/* send ack */
+			kill (slave_pid, SIGUSR2);
+		}
+	} else if (strncmp (msg, GDM_SOP_SESSPID " ",
+		     strlen (GDM_SOP_SESSPID " ")) == 0) {
+		GdmDisplay *d;
+		long slave_pid, pid;
+
+		if (sscanf (msg, GDM_SOP_SESSPID " %ld %ld", &slave_pid, &pid)
+		    != 2)
+			return;
+
+		/* Find out who this slave belongs to */
+		d = gdm_display_lookup (slave_pid);
+
+		if (d != NULL) {
+			d->sesspid = pid;
+			gdm_debug ("Got SESSPID == %ld", (long)pid);
+			/* send ack */
+			kill (slave_pid, SIGUSR2);
+		}
+	} else if (strncmp (msg, GDM_SOP_GREETPID " ",
+		     strlen (GDM_SOP_GREETPID " ")) == 0) {
+		GdmDisplay *d;
+		long slave_pid, pid;
+
+		if (sscanf (msg, GDM_SOP_GREETPID " %ld %ld", &slave_pid, &pid)
+		    != 2)
+			return;
+
+		/* Find out who this slave belongs to */
+		d = gdm_display_lookup (slave_pid);
+
+		if (d != NULL) {
+			d->greetpid = pid;
+			gdm_debug ("Got GREETPID == %ld", (long)pid);
+			/* send ack */
+			kill (slave_pid, SIGUSR2);
+		}
+	} else if (strncmp (msg, GDM_SOP_CHOOSERPID " ",
+		     strlen (GDM_SOP_CHOOSERPID " ")) == 0) {
+		GdmDisplay *d;
+		long slave_pid, pid;
+
+		if (sscanf (msg, GDM_SOP_CHOOSERPID " %ld %ld",
+			    &slave_pid, &pid) != 2)
+			return;
+
+		/* Find out who this slave belongs to */
+		d = gdm_display_lookup (slave_pid);
+
+		if (d != NULL) {
+			d->chooserpid = pid;
+			gdm_debug ("Got CHOOSERPID == %ld", (long)pid);
+			/* send ack */
+			kill (slave_pid, SIGUSR2);
+		}
+	} else if (strncmp (msg, GDM_SOP_LOGGED_IN " ",
+		     strlen (GDM_SOP_LOGGED_IN " ")) == 0) {
+		GdmDisplay *d;
+		long slave_pid;
+		int logged_in;
+		if (sscanf (msg, GDM_SOP_LOGGED_IN " %ld %d", &slave_pid,
+			    &logged_in) != 2)
+			return;
+
+		/* Find out who this slave belongs to */
+		d = gdm_display_lookup (slave_pid);
+
+		if (d != NULL) {
+			d->logged_in = logged_in ? TRUE : FALSE;
+			gdm_debug ("Got logged in == %s",
+				   d->logged_in ? "TRUE" : "FALSE");
+
+			/* if the user just logged out,
+			 * let's see if it's safe to restart */
+			if (gdm_restart_mode &&
+			    ! d->logged_in)
+				gdm_safe_restart ();
+
+			/* send ack */
+			kill (slave_pid, SIGUSR2);
+		}
+	} else if (strcmp (msg, GDM_SOP_SOFT_RESTART) == 0) {
+		gdm_restart_mode = TRUE;
+		gdm_safe_restart ();
+	}
+}
+
+static gboolean
+gdm_slave_socket_handler (GIOChannel *source,
+			  GIOCondition cond,
+			  gpointer data)
+{
+	gchar buf[PIPE_SIZE];
+	char *p;
+	gint len;
+
+	if (cond != G_IO_IN) 
+		return TRUE;
+
+	if (g_io_channel_read (source, buf, sizeof (buf) - 1, &len)
+	    != G_IO_ERROR_NONE)
+		return TRUE;
+
+	if (len <= 0)
+		return TRUE;
+
+	/* null terminate as the string is NOT */
+	buf[len] = '\0';
+
+	p = strtok (buf, "\n");
+	while (p != NULL) {
+		handle_message (p);
+		p = strtok (NULL, "\n");
+	}
+
+	return TRUE;
+}
+
+void
+gdm_fifo_close (void)
+{
+	gchar *fifopath;
+
+	if (fifo_source > 0) {
+		g_source_remove (fifo_source);
+		fifo_source = 0;
+	}
+
+	if (fifofd > 0) {
+		close (fifofd);
+		fifofd = -1;
+	}
+
+	fifopath = g_strconcat (GdmServAuthDir, "/.gdmfifo", NULL);
+	unlink (fifopath);
+	g_free (fifopath);
+}
+
 /* EOF */
