@@ -51,17 +51,17 @@
 #include "filecheck.h"
 #include "auth.h"
 #include "server.h"
+#include "choose.h"
 #include "errorgui.h"
 
 
-static const gchar RCSid[]="$Id: slave.c,v 1.73 2001/06/29 01:48:17 jirka Exp $";
+static const gchar RCSid[]="$Id: slave.c,v 1.79 2001/07/06 11:24:11 jirka Exp $";
 
 /* Some per slave globals */
 static GdmDisplay *d;
 static gchar *login = NULL;
 static sigset_t mask;
 static gboolean greet = FALSE;
-static FILE *greeter = NULL;
 static pid_t last_killed_pid = 0;
 static gboolean do_timed_login = FALSE; /* if this is true,
 					   login the timed login */
@@ -108,7 +108,6 @@ extern gint GdmPingInterval;
 extern gint GdmRetryDelay;
 extern gboolean GdmAllowRoot;
 extern sigset_t sysmask;
-extern gchar *argdelim;
 extern gchar *GdmGlobalFaceDir;
 extern gboolean GdmBrowser;
 
@@ -126,15 +125,19 @@ static void     gdm_slave_session_cleanup (void);
 static void     gdm_slave_alrm_handler (int sig);
 static void     gdm_slave_term_handler (int sig);
 static void     gdm_slave_child_handler (int sig);
-static void     gdm_slave_exit (gint status, const gchar *format, ...);
-static void     gdm_child_exit (gint status, const gchar *format, ...);
+static void     gdm_slave_usr2_handler (int sig);
+static void     gdm_slave_exit (gint status, const gchar *format, ...) G_GNUC_PRINTF (2, 3);
+static void     gdm_child_exit (gint status, const gchar *format, ...) G_GNUC_PRINTF (2, 3);
 static gint     gdm_slave_exec_script (GdmDisplay *d, const gchar *dir,
 				       const char *login, struct passwd *pwent);
 static gchar *  gdm_parse_enriched_login (const gchar *s, GdmDisplay *display);
+static void	gdm_send_logged_in (gboolean logged_in);
+static void	gdm_send_pid (const char *opcode, pid_t pid);
 
 
 /* Yay thread unsafety */
 static gboolean x_error_occured = FALSE;
+static gboolean gdm_got_usr2 = FALSE;
 
 /* ignore handlers */
 static int
@@ -144,19 +147,12 @@ ignore_xerror_handler (Display *disp, XErrorEvent *evt)
 	return 0;
 }
 
-static int
-ignore_xioerror_handler (Display *disp)
-{
-	x_error_occured = TRUE;
-	return 0;
-}
-
 void 
 gdm_slave_start (GdmDisplay *display)
 {  
 	time_t first_time;
 	int death_count;
-	struct sigaction alrm, term, child;
+	struct sigaction alrm, term, child, usr2;
 
 	if (!display)
 		return;
@@ -194,11 +190,21 @@ gdm_slave_start (GdmDisplay *display)
 	if (sigaction (SIGCHLD, &child, NULL) < 0) 
 		gdm_slave_exit (DISPLAY_ABORT, _("gdm_slave_init: Error setting up CHLD signal handler"));
 
+	/* Handle a USR2 which is ack from master that it received a message */
+	usr2.sa_handler = gdm_slave_usr2_handler;
+	usr2.sa_flags = SA_RESTART;
+	sigemptyset (&usr2.sa_mask);
+	sigaddset (&usr2.sa_mask, SIGUSR2);
+
+	if (sigaction (SIGUSR2, &usr2, NULL) < 0)
+		gdm_slave_exit (DISPLAY_ABORT, _("%s: Error setting up USR2 signal handler"), "gdm_slave_init");
+
 	/* The signals we wish to listen to */
 	sigfillset (&mask);
 	sigdelset (&mask, SIGINT);
 	sigdelset (&mask, SIGTERM);
 	sigdelset (&mask, SIGCHLD);
+	sigdelset (&mask, SIGUSR2);
 	if (display->type != TYPE_LOCAL &&
 	    GdmPingInterval > 0) {
 		sigdelset (&mask, SIGALRM);
@@ -238,6 +244,7 @@ gdm_slave_start (GdmDisplay *display)
 			/* Whack the server if we want to restart it next time
 			 * we run gdm_slave_run */
 			gdm_server_stop (display);
+			gdm_send_pid (GDM_SOP_XPID, 0);
 		} else {
 			/* OK about to start again so redo our cookies and reinit
 			 * the server */
@@ -272,18 +279,15 @@ gdm_screen_init (GdmDisplay *display)
 {
 #ifdef HAVE_LIBXINERAMA
 	int (* old_xerror_handler) (Display *, XErrorEvent *);
-	int (* old_xioerror_handler) (Display *);
 	gboolean have_xinerama = FALSE;
 
 	x_error_occured = FALSE;
 	old_xerror_handler = XSetErrorHandler (ignore_xerror_handler);
-	old_xioerror_handler = XSetIOErrorHandler (ignore_xioerror_handler);
 
 	have_xinerama = XineramaIsActive (display->dsp);
 
 	XSync (display->dsp, False);
 	XSetErrorHandler (old_xerror_handler);
-	XSetIOErrorHandler (old_xioerror_handler);
 
 	if (x_error_occured)
 		have_xinerama = FALSE;
@@ -317,6 +321,7 @@ gdm_screen_init (GdmDisplay *display)
 	}
 }
 
+static gboolean do_xfailed_on_xio_error = FALSE;
 
 static void 
 gdm_slave_run (GdmDisplay *display)
@@ -347,8 +352,10 @@ gdm_slave_run (GdmDisplay *display)
     /* if this is local display start a server if one doesn't
      * exist */
     if (d->type == TYPE_LOCAL &&
-	d->servpid <= 0)
+	d->servpid <= 0) {
 	    gdm_server_start (d);
+	    gdm_send_pid (GDM_SOP_XPID, d->servpid);
+    }
     
     gdm_setenv ("XAUTHORITY", d->authfile);
     gdm_setenv ("DISPLAY", d->name);
@@ -356,6 +363,7 @@ gdm_slave_run (GdmDisplay *display)
     /* X error handlers to avoid the default one (i.e. exit (1)) */
     XSetErrorHandler (gdm_slave_xerror_handler);
     XSetIOErrorHandler (gdm_slave_xioerror_handler);
+    do_xfailed_on_xio_error = TRUE;
     
     /* We keep our own (windowless) connection (dsp) open to avoid the
      * X server resetting due to lack of active connections. */
@@ -399,6 +407,10 @@ gdm_slave_run (GdmDisplay *display)
 		    XCloseDisplay (dsp);
     }
 
+    /* OK from now on it's really the user whacking us most likely,
+     * we have already started up well */
+    do_xfailed_on_xio_error = FALSE;
+
     /* If XDMCP setup pinging */
     if (d->type != TYPE_LOCAL &&
 	GdmPingInterval > 0) {
@@ -417,22 +429,31 @@ gdm_slave_run (GdmDisplay *display)
 	       strcmp (ParsedAutomaticLogin, "root") != 0) {
 	    gdm_first_login = FALSE;
 
+	    gdm_send_logged_in (TRUE);
+
 	    setup_automatic_session (d, ParsedAutomaticLogin);
 
 	    gdm_slave_session_start();
 
+	    gdm_send_logged_in (FALSE);
+
 	    gdm_debug ("gdm_slave_run: Automatic login done");
     } else {
 	    if (gdm_first_login)
 		    gdm_first_login = FALSE;
 	    gdm_slave_greeter ();  /* Start the greeter */
 	    gdm_slave_wait_for_login (); /* wait for a password */
+
+	    gdm_send_logged_in (TRUE);
+
 	    if (do_timed_login) {
 		    /* timed out into a timed login */
 		    do_timed_login = FALSE;
 		    setup_automatic_session (d, ParsedTimedLogin);
 	    }
 	    gdm_slave_session_start ();
+
+	    gdm_send_logged_in (FALSE);
     }
 }
 
@@ -458,9 +479,7 @@ gdm_slave_whack_greeter (void)
 		waitpid (d->greetpid, 0, 0); 
 	d->greetpid = 0;
 
-	if (greeter != NULL)
-		fclose (greeter);
-	greeter = NULL;
+	gdm_send_pid (GDM_SOP_GREETPID, 0);
 
 	sigprocmask (SIG_SETMASK, &omask, NULL);
 }
@@ -575,7 +594,7 @@ run_config (GdmDisplay *display, struct passwd *pwent)
 		open("/dev/null", O_RDWR); /* open stderr - fd 2 */
 
 		/* exec the configurator */
-		argv = g_strsplit (GdmConfigurator, " ", MAX_ARGS);
+		argv = gdm_split (GdmConfigurator);
 		if (argv != NULL &&
 		    argv[0] != NULL &&
 		    access (argv[0], X_OK) == 0)
@@ -589,10 +608,9 @@ run_config (GdmDisplay *display, struct passwd *pwent)
 				 "I will attempt to start it from the\n"
 				 "default location."));
 
-		argv = g_strsplit
+		argv = gdm_split
 			(EXPANDED_GDMCONFIGDIR
-			 "/gdmconfig --disable-sound --disable-crash-dialog",
-			 " ", MAX_ARGS);
+			 "/gdmconfig --disable-sound --disable-crash-dialog");
 		if (access (argv[0], X_OK) == 0)
 			execv (argv[0], argv);
 
@@ -697,6 +715,8 @@ gdm_slave_wait_for_login (void)
 				continue;
 			}
 
+			gdm_send_logged_in (TRUE);
+
 			/* disable the login screen, we don't want people to
 			 * log in in the meantime */
 			gdm_slave_greeter_ctl_no_ret (GDM_DISABLE, "");
@@ -708,6 +728,8 @@ gdm_slave_wait_for_login (void)
 
 			gdm_verify_cleanup ();
 
+			gdm_send_logged_in (FALSE);
+
 			gdm_slave_greeter_ctl_no_ret (GDM_ENABLE, "");
 			gdm_slave_greeter_ctl_no_ret (GDM_RESETOK, "");
 			continue;
@@ -1056,7 +1078,7 @@ gdm_slave_greeter (void)
 				 "configure the X server."));
 	}
 
-	argv = g_strsplit (GdmGreeter, argdelim, MAX_ARGS);
+	argv = gdm_split (GdmGreeter);
 	execv (argv[0], argv);
 
 	gdm_error (_("gdm_slave_greeter: Cannot start greeter trying default: %s"),
@@ -1065,9 +1087,8 @@ gdm_slave_greeter (void)
 
 	gdm_setenv ("GDM_WHACKED_GREETER_CONFIG", "true");
 
-	argv = g_strsplit (EXPANDED_BINDIR
-			   "/gdmlogin --disable-sound --disable-crash-dialog",
-			   argdelim, MAX_ARGS);
+	argv = gdm_split (EXPANDED_BINDIR
+			  "/gdmlogin --disable-sound --disable-crash-dialog");
 	execv (argv[0], argv);
 
 	gdm_error_box (d,
@@ -1098,28 +1119,106 @@ gdm_slave_greeter (void)
 	fcntl(pipe1[1], F_SETFD, fcntl(pipe1[1], F_GETFD, 0) | FD_CLOEXEC);
 	fcntl(pipe2[0], F_SETFD, fcntl(pipe2[0], F_GETFD, 0) | FD_CLOEXEC);
 
+	/* flush our input before we change the piping */
+	fflush (stdin);
+
 	if (pipe1[1] != STDOUT_FILENO) 
 	    dup2 (pipe1[1], STDOUT_FILENO);
 	
 	if (pipe2[0] != STDIN_FILENO) 
 	    dup2 (pipe2[0], STDIN_FILENO);
-	
-	greeter = fdopen (STDIN_FILENO, "r");
+
+	close (pipe1[1]);
+	close (pipe2[0]);
 	
 	gdm_debug ("gdm_slave_greeter: Greeter on pid %d", d->greetpid);
 
+	gdm_send_pid (GDM_SOP_GREETPID, d->greetpid);
+
 	run_pictures (); /* Append pictures to greeter if browsing is on */
 	break;
     }
 }
 
+static void
+gdm_send_pid (const char *opcode, pid_t pid)
+{
+	char *msg;
+	int fd;
+	char *fifopath;
+
+	gdm_debug ("Sending %s == %ld for slave %ld",
+		   opcode,
+		   (long)pid,
+		   (long)getpid ());
+
+	gdm_got_usr2 = FALSE;
+
+	fifopath = g_strconcat (GdmServAuthDir, "/.gdmfifo", NULL);
+
+	fd = open (fifopath, O_WRONLY);
+	/* eek */
+	if (fd < 0) {
+		gdm_error (_("%s: Can't open fifo!"), "gdm_send_pid");
+		return;
+	}
+
+	msg = g_strdup_printf ("\n%s %ld %ld\n", opcode,
+			       (long)getpid (), (long)pid);
+
+	write (fd, msg, strlen (msg));
+
+	g_free (msg);
+
+	close (fd);
+
+	if ( ! gdm_got_usr2)
+		sleep (10);
+}
+
+
+static void
+gdm_send_logged_in (gboolean logged_in)
+{
+	char *msg;
+	int fd;
+	char *fifopath;
+
+	gdm_debug ("Sending logged_in == %s for slave %ld",
+		   logged_in ? "TRUE" : "FALSE",
+		   (long)getpid ());
+
+	gdm_got_usr2 = FALSE;
+
+	fifopath = g_strconcat (GdmServAuthDir, "/.gdmfifo", NULL);
+
+	fd = open (fifopath, O_WRONLY);
+	/* eek */
+	if (fd < 0) {
+		gdm_error (_("%s: Can't open fifo!"), "gdm_send_logged_in");
+		return;
+	}
+
+	msg = g_strdup_printf ("\n%s %ld %d\n", GDM_SOP_LOGGED_IN,
+			       (long)getpid (), (int)logged_in);
+
+	write (fd, msg, strlen (msg));
+
+	g_free (msg);
+
+	close (fd);
+
+	if ( ! gdm_got_usr2)
+		sleep (10);
+}
+
 static void
 send_chosen_host (GdmDisplay *disp, const char *hostname)
 {
+	char *msg;
 	int fd;
 	char *fifopath;
 	struct hostent *host;
-	char *buf;
 
 	host = gethostbyname (hostname);
 
@@ -1131,18 +1230,23 @@ send_chosen_host (GdmDisplay *disp, const char *hostname)
 	gdm_debug ("Sending chosen host address (%s) %s",
 		   hostname, inet_ntoa (*(struct in_addr *)host->h_addr_list[0]));
 
-	fifopath = g_strconcat (GdmServAuthDir, "/.gdmchooser", NULL);
+	fifopath = g_strconcat (GdmServAuthDir, "/.gdmfifo", NULL);
 
 	fd = open (fifopath, O_WRONLY);
 
-	/* send char: STX, int: indirect display id, in_addr: address */
-	buf = g_new (char, 1 + sizeof (int) + sizeof (struct in_addr));
-	buf[0] = STX;
-	memcpy (&buf[1], &disp->indirect_id, sizeof (int));
-	memcpy (&buf[1 + sizeof (int)],
-		(struct in_addr *)host->h_addr_list[0],
-		sizeof (struct in_addr));
-	write (fd, buf, 1 + sizeof (int) + sizeof (struct in_addr));
+	/* eek */
+	if (fd < 0) {
+		gdm_error (_("%s: Can't open fifo!"), "send_chosen_host");
+		return;
+	}
+
+	msg = g_strdup_printf ("\n%s %d %s\n", GDM_SOP_CHOSEN,
+			       disp->indirect_id,
+			       inet_ntoa (*((struct in_addr *)host->h_addr_list[0])));
+
+	write (fd, msg, strlen (msg));
+
+	g_free (msg);
 
 	close (fd);
 }
@@ -1169,7 +1273,7 @@ gdm_slave_chooser (void)
 	/* Fork. Parent is gdmslave, child is greeter process. */
 	last_killed_pid = 0; /* race condition wrapper,
 			      * it could be that we recieve sigchld before
-			      * we can set greetpid.  eek! */
+			      * we can set chooserpid.  eek! */
 	switch (d->chooserpid = fork()) {
 
 	case 0:
@@ -1219,7 +1323,7 @@ gdm_slave_chooser (void)
 		gdm_setenv ("PATH", GdmDefaultPath);
 		gdm_setenv ("RUNNING_UNDER_GDM", "true");
 
-		argv = g_strsplit (GdmChooser, argdelim, MAX_ARGS);
+		argv = gdm_split (GdmChooser);
 		execv (argv[0], argv);
 
 		gdm_error_box (d,
@@ -1243,6 +1347,8 @@ gdm_slave_chooser (void)
 		}
 
 		gdm_debug ("gdm_slave_chooser: Chooser on pid %d", d->chooserpid);
+		gdm_send_pid (GDM_SOP_CHOOSERPID, d->chooserpid);
+
 		close (p[1]);
 
 		fcntl(p[0], F_SETFD, fcntl(p[0], F_GETFD, 0) | FD_CLOEXEC);
@@ -1250,6 +1356,8 @@ gdm_slave_chooser (void)
 		/* wait for the chooser to die */
 		waitpid (d->chooserpid, 0, 0);
 
+		gdm_send_pid (GDM_SOP_CHOOSERPID, 0);
+
 		bytes = read (p[0], buf, sizeof(buf)-1);
 		if (bytes > 0) {
 			if (buf[bytes-1] == '\n')
@@ -1258,8 +1366,12 @@ gdm_slave_chooser (void)
 				buf[bytes] ='\0';
 			send_chosen_host (d, buf);
 
+			close (p[0]);
+
 			_exit (DISPLAY_CHOSEN);
 		}
+
+		close (p[0]);
 		break;
 	}
 }
@@ -1510,6 +1622,27 @@ unaliaslang (const char *origlang)
 
 }
 
+static char *
+dequote (const char *in)
+{
+	GString *str;
+	char *out;
+	const char *p;
+
+	str = g_string_new (NULL);
+
+	for (p = in; *p != '\0'; p++) {
+		if (*p == '\'')
+			g_string_append (str, "'\\''");
+		else
+			g_string_append_c (str, *p);
+	}
+
+	out = str->str;
+	g_string_free (str, FALSE);
+	return out;
+}
+
 static void
 gdm_slave_session_start (void)
 {
@@ -1877,10 +2010,15 @@ gdm_slave_session_start (void)
 	} 
 
 	if (sesspath == NULL) {
-		if (GdmSessDir != NULL)
-			sesspath = g_strconcat (GdmSessDir, "/", session, NULL);
-		else
-			sesspath = g_strdup ("/Eeeeek! Eeeeek!");
+		if (GdmSessDir != NULL) {
+			sesspath = g_strconcat
+				("'", GdmSessDir, "/",
+				 dequote (session), "'", NULL);
+			sessexec = g_strconcat (GdmSessDir, "/",
+						session, NULL);
+		} else {
+			sesspath = g_strdup ("'/Eeeeek! Eeeeek!'");
+		}
 	}
 	
 	gdm_debug (_("Running %s for %s on %s"),
@@ -1934,6 +2072,7 @@ gdm_slave_session_start (void)
     g_free (gnome_session);
 
     sesspid = d->sesspid;
+    gdm_send_pid (GDM_SOP_SESSPID, sesspid);
 
     /* Wait for the user's session to exit, but by this time the
      * session might have ended, so check for 0 */
@@ -1960,6 +2099,8 @@ gdm_slave_session_stop (pid_t sesspid)
     seteuid (0);
     setegid (0);
 
+    gdm_send_pid (GDM_SOP_SESSPID, 0);
+
     gdm_debug ("gdm_slave_session_stop: %s on %s", local_login, d->name);
 
     if (sesspid > 0)
@@ -2000,7 +2141,6 @@ gdm_slave_session_cleanup (void)
     
     /* things are going to be killed, so ignore errors */
     XSetErrorHandler (ignore_xerror_handler);
-    XSetIOErrorHandler (ignore_xioerror_handler);
 
     /* Cleanup */
     gdm_debug ("gdm_slave_session_cleanup: Severing connection");
@@ -2015,7 +2155,7 @@ gdm_slave_term_handler (int sig)
 {
 	sigset_t tmask, omask;
 
-	gdm_debug ("gdm_slave_term_handler: %s got TERM signal", d->name);
+	gdm_debug ("gdm_slave_term_handler: %s got TERM/INT signal", d->name);
 
 	/* just for paranoia's sake */
 	seteuid (0);
@@ -2130,6 +2270,14 @@ gdm_slave_child_handler (int sig)
     }
 }
 
+static void
+gdm_slave_usr2_handler (int sig)
+{
+	gdm_debug ("gdm_slave_usr2_handler: %s got USR2 signal", d->name);
+	
+	gdm_got_usr2 = TRUE;
+}
+
 /* Minor X faults */
 static gint
 gdm_slave_xerror_handler (Display *disp, XErrorEvent *evt)
@@ -2138,7 +2286,6 @@ gdm_slave_xerror_handler (Display *disp, XErrorEvent *evt)
     return (0);
 }
 
-
 /* We respond to fatal errors by restarting the display */
 static gint
 gdm_slave_xioerror_handler (Display *disp)
@@ -2170,9 +2317,10 @@ gdm_slave_xioerror_handler (Display *disp)
 	gdm_server_stop (d);
 	gdm_verify_cleanup ();
 
-	sigprocmask (SIG_SETMASK, &omask, NULL);
-
-	_exit (DISPLAY_XFAILED);
+	if (do_xfailed_on_xio_error)
+		_exit (DISPLAY_XFAILED);
+	else
+		_exit (DISPLAY_REMANAGE);
 }
 
 char * 
@@ -2181,9 +2329,6 @@ gdm_slave_greeter_ctl (char cmd, const char *str)
     gchar buf[FIELD_SIZE];
     guchar c;
 
-    if (greeter == NULL)
-	    return NULL;
-
     if (str)
 	g_print ("%c%c%s\n", STX, cmd, str);
     else
@@ -2191,16 +2336,16 @@ gdm_slave_greeter_ctl (char cmd, const char *str)
 
     /* Skip random junk that might have accumulated */
     do {
-	    c = getc (greeter);
+	    c = getc (stdin);
     } while (c && c != STX);
     
-    if (fgets (buf, FIELD_SIZE-1, greeter) == NULL) {
+    if (fgets (buf, FIELD_SIZE-1, stdin) == NULL) {
 	    /* things don't seem well with the greeter, it probably died */
 	    return NULL;
     }
 
     /* don't forget to flush */
-    fflush (greeter);
+    fflush (stdin);
     
     if (buf[0] != '\0') {
 	    int len = strlen (buf);
@@ -2323,7 +2468,7 @@ gdm_slave_exec_script (GdmDisplay *d, const gchar *dir, const char *login,
         gdm_setenv ("DISPLAY", d->name);
 	gdm_setenv ("PATH", GdmRootPath);
 	gdm_unsetenv ("MAIL");
-	argv = g_strsplit (scr, argdelim, MAX_ARGS);
+	argv = gdm_split (scr);
 	execv (argv[0], argv);
 	syslog (LOG_ERR, _("gdm_slave_exec_script: Failed starting: %s"), scr);
 	_exit (EXIT_SUCCESS);
@@ -2468,7 +2613,7 @@ gdm_parse_enriched_login (const gchar *s, GdmDisplay *display)
 	  gdm_setenv ("PATH", GdmRootPath);
 	  gdm_unsetenv ("MAIL");
 
-	  argv = g_strsplit (str->str, argdelim, MAX_ARGS);
+	  argv = gdm_split (str->str);
 	  execv (argv[0], argv);
 	  syslog (LOG_ERR, _("gdm_parse_enriched_login: Failed executing: %s"),
 		  str->str);
