@@ -32,14 +32,12 @@
 #include <ctype.h>
 #include <signal.h>
 #include <dirent.h>
+#include <locale.h>
 #include <gdk/gdkx.h>
 #include <X11/X.h>
 #include <X11/Xlib.h>
 #include <pwd.h>
 #include <sys/utsname.h>
-#ifdef HAVE_LIBXINERAMA
-#include <X11/extensions/Xinerama.h>
-#endif
 #include "gdmlogin.h"
 #include "gdm.h"
 #include "gdmwm.h"
@@ -50,7 +48,7 @@
  * within the protocol */
 static gboolean DOING_GDM_DEVELOPMENT = FALSE;
 
-static const gchar RCSid[]="$Id: gdmlogin.c,v 1.83 2001/05/31 18:37:45 hp Exp $";
+static const gchar RCSid[]="$Id: gdmlogin.c,v 1.96 2001/07/06 08:53:57 jirka Exp $";
 
 /* Some strings that are in other files that we may want to
  * translate.  This is not actually used anywhere, it's just
@@ -78,7 +76,11 @@ static gint  GdmIconMaxHeight;
 static gint  GdmIconMaxWidth;
 static gboolean GdmQuiver;
 static gboolean GdmSystemMenu;
+static gchar *GdmHalt;
+static gchar *GdmReboot;
+static gchar *GdmSuspend;
 static gboolean GdmConfigAvailable;
+static gchar *GdmConfigurator;
 static gint GdmXineramaScreen;
 static gchar *GdmLogo;
 static gchar *GdmWelcome;
@@ -125,17 +127,11 @@ static gboolean remember_gnome_session = FALSE;
 static GtkWidget *icon_win = NULL;
 static GtkWidget *sessmenu;
 static GtkWidget *langmenu;
-static GdkRectangle *allscreens;
-static int screens = 0;
 static GtkTooltips *tooltips;
 
 static gboolean login_is_local = FALSE;
 static gboolean used_defaults = FALSE;
 
-
-
-GdkRectangle screen; /* this is an extern since it's used in gdmwm as well */
-
 static GnomeIconList *browser;
 static GdkImlibImage *defface;
 
@@ -165,10 +161,14 @@ static pid_t backgroundpid = 0;
 
 static guint timed_handler_id = 0;
 
+static char *selected_browser_user = NULL;
+
 /* This is true if session dir doesn't exist or is whacked out
  * in some way or another */
 static gboolean session_dir_whacked_out = FALSE;
 
+static void gdm_login_abort (const gchar *format, ...) G_GNUC_PRINTF (1, 2);
+
 /*
  * Timed Login: Timer
  */
@@ -276,14 +276,14 @@ set_screen_pos (GtkWidget *widget, int x, int y)
 	g_return_if_fail (widget != NULL);
 	g_return_if_fail (GTK_IS_WIDGET (widget));
 
-	if (x < screen.x)
-		x = screen.x;
-	if (y < screen.y)
-		y = screen.y;
-	if (x > screen.x + screen.width - widget->allocation.width)
-		x = screen.x + screen.width - widget->allocation.width;
-	if (y > screen.y + screen.height - widget->allocation.height)
-		y = screen.y + screen.height - widget->allocation.height;
+	if (x < gdm_wm_screen.x)
+		x = gdm_wm_screen.x;
+	if (y < gdm_wm_screen.y)
+		y = gdm_wm_screen.y;
+	if (x > gdm_wm_screen.x + gdm_wm_screen.width - widget->allocation.width)
+		x = gdm_wm_screen.x + gdm_wm_screen.width - widget->allocation.width;
+	if (y > gdm_wm_screen.y + gdm_wm_screen.height - widget->allocation.height)
+		y = gdm_wm_screen.y + gdm_wm_screen.height - widget->allocation.height;
 
 	gtk_widget_set_uposition (widget, x, y);
 }
@@ -360,15 +360,14 @@ within_rect (GdkRectangle *rect, int x, int y)
 static void
 set_screen_to_pos (int x, int y)
 {
-	if ( ! within_rect (&screen, x, y)) {
+	if ( ! within_rect (&gdm_wm_screen, x, y)) {
 		int i;
-		/* If not within screen boundaries,
+		/* If not within gdm_wm_screen boundaries,
 		 * maybe we want to switch xinerama
 		 * screen */
-		for (i = 0; i < screens; i++) {
-			if (within_rect (&allscreens[i], x, y)) {
-				GdmXineramaScreen = i;
-				screen = allscreens[i];
+		for (i = 0; i < gdm_wm_screens; i++) {
+			if (within_rect (&gdm_wm_allscreens[i], x, y)) {
+				gdm_wm_set_screen (i);
 				break;
 			}
 		}
@@ -450,8 +449,8 @@ gdm_login_iconify_handler (GtkWidget *widget, gpointer data)
 
     gtk_widget_show (GTK_WIDGET (icon_win));
 
-    rw = screen.width;
-    rh = screen.height;
+    rw = gdm_wm_screen.width;
+    rh = gdm_wm_screen.height;
 
     set_screen_pos (GTK_WIDGET (icon_win), rw-iw, rh-ih);
 }
@@ -465,7 +464,7 @@ gdm_login_abort (const gchar *format, ...)
 
     if (!format) {
 	kill_thingies ();
-	exit (DISPLAY_ABORT);
+	_exit (DISPLAY_ABORT);
     }
 
     va_start (args, format);
@@ -476,7 +475,7 @@ gdm_login_abort (const gchar *format, ...)
     closelog();
 
     kill_thingies ();
-    exit (DISPLAY_ABORT);
+    _exit (DISPLAY_ABORT);
 }
 
 
@@ -587,25 +586,6 @@ gdm_parse_enriched_string (const gchar *s)
     return buffer;
 }
 
-static void
-gdm_center_window (GtkWindow *cw) 
-{
-	GtkRequisition req;
-        gint x, y;
-
-	gtk_widget_size_request (GTK_WIDGET (cw), &req);
-
-	x = screen.x + (screen.width - req.width)/2;
-	y = screen.y + (screen.height - req.height)/2;	
-
-	if (x < screen.x)
-		x = screen.x;
-	if (y < screen.y)
-		y = screen.y;
-
- 	gtk_widget_set_uposition (GTK_WIDGET (cw), x, y);	
-}
-
 static gboolean
 gdm_login_query (const gchar *msg)
 {
@@ -628,7 +608,7 @@ gdm_login_query (const gchar *msg)
 			    &req);
 
 	gtk_window_set_modal (GTK_WINDOW (req), TRUE);
-	gdm_center_window (GTK_WINDOW (req));
+	gdm_wm_center_window (GTK_WINDOW (req));
 
 	gdm_wm_no_login_focus_push ();
 	ret = gnome_dialog_run (GNOME_DIALOG (req));
@@ -657,7 +637,7 @@ gdm_run_command (const char *command)
 			   "You likely won't be able to log "
 			   "in either."));
 		gtk_widget_show_all (dialog);
-		gdm_center_window (GTK_WINDOW (dialog));
+		gdm_wm_center_window (GTK_WINDOW (dialog));
 		gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
 
 		gnome_dialog_run_and_close (GNOME_DIALOG (dialog));
@@ -675,7 +655,7 @@ gdm_run_command (const char *command)
 		open("/dev/null", O_RDWR); /* open stdout - fd 1 */
 		open("/dev/null", O_RDWR); /* open stderr - fd 2 */
 
-		argv = g_strsplit (command, " ", MAX_ARGS);	
+		argv = gdm_split (command);
 		execv (argv[0], argv);
 		/*ingore errors, this is irrelevant */
 		_exit (0);
@@ -697,33 +677,39 @@ gdm_run_gdmconfig (GtkWidget *w, gpointer data)
 	g_print ("%c%c%c\n", STX, BEL, GDM_INTERRUPT_CONFIGURE);
 }
 
-static gboolean
+static void
 gdm_login_reboot_handler (void)
 {
-    if (gdm_login_query (_("Are you sure you want to reboot the machine?"))) {
-	closelog();
+	if (gdm_login_query (_("Are you sure you want to reboot the machine?"))) {
+		closelog();
 
-        kill_thingies ();
-	exit (DISPLAY_REBOOT);
-    }
-
-    return (TRUE);
+		kill_thingies ();
+		_exit (DISPLAY_REBOOT);
+	}
 }
 
 
-static gboolean
+static void
 gdm_login_halt_handler (void)
 {
-    if (gdm_login_query (_("Are you sure you want to halt the machine?"))) {
-	closelog();
+	if (gdm_login_query (_("Are you sure you want to halt the machine?"))) {
+		closelog();
 
-        kill_thingies ();
-	exit (DISPLAY_HALT);
-    }
-
-    return (TRUE);
+		kill_thingies ();
+		_exit (DISPLAY_HALT);
+	}
 }
 
+static void
+gdm_login_suspend_handler (void)
+{
+	if (gdm_login_query (_("Are you sure you want to suspend the machine?"))) {
+		closelog();
+
+		kill_thingies ();
+		_exit (DISPLAY_SUSPEND);
+	}
+}
 
 static void 
 gdm_login_parse_config (void)
@@ -745,7 +731,11 @@ gdm_login_parse_config (void)
     GdmIcon = gnome_config_get_string (GDM_KEY_ICON);
     GdmQuiver = gnome_config_get_bool (GDM_KEY_QUIVER);
     GdmSystemMenu = gnome_config_get_bool (GDM_KEY_SYSMENU);
+    GdmHalt = gnome_config_get_string (GDM_KEY_HALT);
+    GdmReboot = gnome_config_get_string (GDM_KEY_REBOOT);
+    GdmSuspend = gnome_config_get_string (GDM_KEY_SUSPEND);
     GdmConfigAvailable = gnome_config_get_bool (GDM_KEY_CONFIG_AVAILABLE);
+    GdmConfigurator = gnome_config_get_string (GDM_KEY_CONFIGURATOR);
     GdmTitleBar = gnome_config_get_bool (GDM_KEY_TITLE_BAR);
     GdmLocaleFile = gnome_config_get_string (GDM_KEY_LOCFILE);
     GdmDefaultLocale = gnome_config_get_string (GDM_KEY_LOCALE);
@@ -777,18 +767,27 @@ gdm_login_parse_config (void)
     GdmShowGnomeChooserSession = gnome_config_get_bool (GDM_KEY_SHOW_GNOME_CHOOSER);
     
     GdmTimedLoginEnable = gnome_config_get_bool (GDM_KEY_TIMED_LOGIN_ENABLE);
+
+    /* Note: TimedLogin here is not gotten out of the config
+     * but from the daemon since it's been munged on by the daemon a bit
+     * already maybe */
     if (GdmTimedLoginEnable) {
-	    GdmTimedLogin = gnome_config_get_string (GDM_KEY_TIMED_LOGIN);
+	    GdmTimedLogin = g_getenv("GDM_TIMED_LOGIN_OK");
+            if (gdm_string_empty (GdmTimedLogin)) {
+	      g_free (GdmTimedLogin);
+	      GdmTimedLogin = NULL;
+	    }
+
 	    GdmTimedLoginDelay =
 		    gnome_config_get_int (GDM_KEY_TIMED_LOGIN_DELAY);
-	    if (GdmTimedLoginDelay < 10) {
+	    if (GdmTimedLoginDelay < 5) {
 		    syslog (LOG_WARNING,
-			    _("TimedLoginDelay was less then 10.  I'll just use 10."));
-		    GdmTimedLoginDelay = 10;
+			    _("TimedLoginDelay was less then 5.  I'll just use 5."));
+		    GdmTimedLoginDelay = 5;
 	    }
     } else {
 	    GdmTimedLogin = NULL;
-	    GdmTimedLoginDelay = 10;
+	    GdmTimedLoginDelay = 5;
     }
 
     gnome_config_pop_prefix();
@@ -813,12 +812,6 @@ gdm_login_parse_config (void)
     } else {
 	    login_is_local = TRUE;
     }
-
-    /* Disable timed login stuff if it's not ok for this display */
-    if (gdm_string_empty (g_getenv ("GDM_TIMED_LOGIN_OK"))) {
-	    g_free (GdmTimedLogin);
-	    GdmTimedLogin = NULL;
-    }
 }
 
 
@@ -936,7 +929,8 @@ gdm_login_language_lookup (const gchar* savedlang)
     /* Don't save language unless told otherwise */
     savelang = FALSE;
 
-    gtk_widget_set_sensitive (GTK_WIDGET (langmenu), FALSE);
+    if (langmenu != NULL)
+	    gtk_widget_set_sensitive (GTK_WIDGET (langmenu), FALSE);
 
     /* Previously saved language not found in ~user/.gnome/gdm */
     if (gdm_string_empty (savedlang)) {
@@ -1007,9 +1001,9 @@ dance (gpointer data)
 	static int height = -1;
 
 	if (width == -1)
-		width = screen.width;
+		width = gdm_wm_screen.width;
 	if (height == -1)
-		height = screen.height;
+		height = gdm_wm_screen.height;
 
 	if (login == NULL ||
 	    login->window == NULL) {
@@ -1023,8 +1017,8 @@ dance (gpointer data)
 	t1 += 0.03 + (rand () % 10) / 500.0;
 	t2 += 0.03 + (rand () % 10) / 500.0;
 
-	x = screen.x + (width / 2) + (width / 5) * xm;
-	y = screen.y + (height / 2) + (height / 5) * ym;
+	x = gdm_wm_screen.x + (width / 2) + (width / 5) * xm;
+	y = gdm_wm_screen.y + (height / 2) + (height / 5) * ym;
 
 	set_screen_pos (login,
 			x - login->allocation.width / 2,
@@ -1054,7 +1048,7 @@ evil (const char *user)
 		gtk_timeout_remove (dance_handler);
 		dance_handler = 0;
 		GdmLockPosition = old_lock;
-		gdm_center_window (GTK_WINDOW (login));
+		gdm_wm_center_window (GTK_WINDOW (login));
 		gtk_entry_set_text (GTK_ENTRY (entry), "");
 		return TRUE;
 				 /* do not translate */
@@ -1705,7 +1699,7 @@ get_gnome_session (const char *sess_string)
 
 	gtk_widget_show_all (d);
 
-	gdm_center_window (GTK_WINDOW (d));
+	gdm_wm_center_window (GTK_WINDOW (d));
 	gdm_wm_no_login_focus_push ();
 	gnome_dialog_run (GNOME_DIALOG (d));
 	gdm_wm_no_login_focus_pop ();
@@ -1966,7 +1960,7 @@ gdm_login_ctrl_handler (GIOChannel *source, GIOCondition cond, gint fd)
 					   GNOME_STOCK_BUTTON_OK,
 					   NULL);
 		gtk_window_set_modal (GTK_WINDOW (d), TRUE);
-		gdm_center_window (GTK_WINDOW (d));
+		gdm_wm_center_window (GTK_WINDOW (d));
 
 		gdm_wm_no_login_focus_push ();
 		gnome_dialog_run (GNOME_DIALOG (d));
@@ -1975,9 +1969,12 @@ gdm_login_ctrl_handler (GIOChannel *source, GIOCondition cond, gint fd)
 
 	kill_thingies ();
 
+	gdk_flush ();
+
 	g_print ("%c\n", STX);
 
-	gtk_main_quit ();
+	/* screw gtk_main_quit, we want to make sure we definately die */
+	_exit (DISPLAY_SUCCESS);
 	break;
 
     case GDM_GNOMESESS:
@@ -2107,16 +2104,22 @@ gdm_login_browser_select (GtkWidget *widget, gint selected, GdkEvent *event)
     case GDK_BUTTON_RELEASE:
 	user = g_list_nth_data (users, selected);
 
-	if (user && user->login)
-	    gtk_entry_set_text (GTK_ENTRY (entry), user->login);
+	if (user && user->login) {
+		gtk_entry_set_text (GTK_ENTRY (entry), user->login);
+		g_free (selected_browser_user);
+		selected_browser_user = g_strdup (user->login);
+	}
 
 	break;
 
     case GDK_2BUTTON_PRESS:
 	user = g_list_nth_data (users, selected);
 
-	if (user && user->login)
-	    gtk_entry_set_text (GTK_ENTRY (entry), user->login);
+	if (user && user->login) {
+		gtk_entry_set_text (GTK_ENTRY (entry), user->login);
+		g_free (selected_browser_user);
+		selected_browser_user = g_strdup (user->login);
+	}
 
 	if (curuser == NULL)
 	    curuser = g_strdup (gtk_entry_get_text (GTK_ENTRY (entry)));
@@ -2146,6 +2149,8 @@ gdm_login_browser_unselect (GtkWidget *widget, gint selected, GdkEvent *event)
     case GDK_BUTTON_PRESS:
     case GDK_BUTTON_RELEASE:
 	gtk_entry_set_text (GTK_ENTRY (entry), "");
+	g_free (selected_browser_user);
+	selected_browser_user = NULL;
 	break;
 	
     default:
@@ -2419,6 +2424,88 @@ clock_state_changed (GtkWidget *clock)
 	clock->state = login->state;
 }
 
+/* doesn't check for executability, just for existance */
+static gboolean
+bin_exists (const char *command)
+{
+	char **argv;
+
+	if (gdm_string_empty (command))
+		return FALSE;
+
+	/* Note, check only for existance, not for executability */
+	argv = gdm_split (command);
+	if (argv != NULL &&
+	    argv[0] != NULL &&
+	    access (argv[0], F_OK) == 0) {
+		g_strfreev (argv);
+		return TRUE;
+	} else {
+		g_strfreev (argv);
+		return FALSE;
+	}
+}
+
+static gboolean
+window_browser_event (GtkWidget *window, GdkEvent *event, gpointer data)
+{
+	switch (event->type) {
+	case GDK_KEY_PRESS:
+		if ((event->key.state & GDK_CONTROL_MASK) &&
+		    (event->key.keyval == GDK_f ||
+		     event->key.keyval == GDK_F) &&
+		    selected_browser_user != NULL) {
+			GtkWidget *d, *less;
+			char *command;
+			d = gnome_dialog_new (_("Finger"),
+					      GNOME_STOCK_BUTTON_OK,
+					      NULL);
+			less = gnome_less_new ();
+			gtk_widget_show (less);
+			gtk_box_pack_start (GTK_BOX (GNOME_DIALOG (d)->vbox),
+					    less,
+					    TRUE,
+					    TRUE,
+					    0);
+
+			/* hack to make this be the size of a terminal */
+			gnome_less_set_fixed_font (GNOME_LESS (less), TRUE);
+			{
+				int i;
+				char buf[82];
+				GtkWidget *text = GTK_WIDGET (GNOME_LESS (less)->text);
+				GdkFont *font = GNOME_LESS (less)->font;
+				for (i = 0; i < 81; i++)
+					buf[i] = 'X';
+				buf[i] = '\0';
+				gtk_widget_set_usize
+					(text,
+					 gdk_string_width (font, buf) + 30,
+					 gdk_string_height (font, buf)*24+30);
+			}
+
+			command = g_strconcat ("finger ",
+					       selected_browser_user,
+					       NULL);
+			gnome_less_show_command (GNOME_LESS (less), command);
+
+			gtk_widget_grab_focus (GTK_WIDGET (less));
+
+			gtk_window_set_modal (GTK_WINDOW (d), TRUE);
+			gdm_wm_center_window (GTK_WINDOW (d));
+
+			gdm_wm_no_login_focus_push ();
+			gnome_dialog_run_and_close (GNOME_DIALOG (d));
+			gdm_wm_no_login_focus_pop ();
+		}
+		break;
+	default:
+		break;
+	}
+
+	return FALSE;
+}
+
 static void
 gdm_login_gui_init (void)
 {
@@ -2440,6 +2527,11 @@ gdm_login_gui_init (void)
     gtk_object_set_data_full (GTK_OBJECT (login), "login", login,
 			      (GtkDestroyNotify) gtk_widget_unref);
     gtk_window_set_title (GTK_WINDOW (login), _("GDM Login"));
+    /* connect for fingering */
+    if (GdmBrowser)
+	    gtk_signal_connect (GTK_OBJECT (login), "event",
+				GTK_SIGNAL_FUNC (window_browser_event),
+				NULL);
 
     accel = gtk_accel_group_new ();
     gtk_window_add_accel_group (GTK_WINDOW (login), accel);
@@ -2500,47 +2592,83 @@ gdm_login_gui_init (void)
     }
 
     if (GdmSystemMenu) {
+        gboolean got_anything = FALSE;
+
 	menu = gtk_menu_new();
-        if (GdmConfigAvailable) {
-	   item = gtk_menu_item_new_with_label (_("Configure..."));
-	   gtk_menu_append (GTK_MENU (menu), item);
-	   gtk_signal_connect (GTK_OBJECT (item), "activate",
-			       GTK_SIGNAL_FUNC (gdm_run_gdmconfig),
-			       NULL);
-	   gtk_widget_show (item);
-	   gtk_tooltips_set_tip (tooltips, GTK_WIDGET (item),
-				 _("Configure GDM (this login manager). "
-				   "This will require the root password."),
-				 NULL);
+	if (GdmConfigAvailable &&
+	    bin_exists (GdmConfigurator)) {
+		item = gtk_menu_item_new_with_label (_("Configure..."));
+		gtk_menu_append (GTK_MENU (menu), item);
+		gtk_signal_connect (GTK_OBJECT (item), "activate",
+				    GTK_SIGNAL_FUNC (gdm_run_gdmconfig),
+				    NULL);
+		gtk_widget_show (item);
+		gtk_tooltips_set_tip (tooltips, GTK_WIDGET (item),
+				      _("Configure GDM (this login manager). "
+					"This will require the root password."),
+				      NULL);
+		got_anything = TRUE;
 	}
 
-	item = gtk_menu_item_new_with_label (_("Reboot..."));
-	gtk_menu_append (GTK_MENU (menu), item);
-	gtk_signal_connect (GTK_OBJECT (item), "activate",
-			    GTK_SIGNAL_FUNC (gdm_login_reboot_handler), 
-			    NULL);
-	gtk_widget_show (GTK_WIDGET (item));
-	gtk_tooltips_set_tip (tooltips, GTK_WIDGET (item),
-			      _("Reboot your computer"),
-			      NULL);
+	if (bin_exists (GdmReboot)) {
+		item = gtk_menu_item_new_with_label (_("Reboot..."));
+		gtk_menu_append (GTK_MENU (menu), item);
+		gtk_signal_connect (GTK_OBJECT (item), "activate",
+				    GTK_SIGNAL_FUNC (gdm_login_reboot_handler), 
+				    NULL);
+		gtk_widget_show (GTK_WIDGET (item));
+		gtk_tooltips_set_tip (tooltips, GTK_WIDGET (item),
+				      _("Reboot your computer"),
+				      NULL);
+		got_anything = TRUE;
+	}
 	
-	item = gtk_menu_item_new_with_label (_("Halt..."));
-	gtk_menu_append (GTK_MENU (menu), item);
-	gtk_signal_connect (GTK_OBJECT (item), "activate",
-			   GTK_SIGNAL_FUNC (gdm_login_halt_handler), 
-			    NULL);
-	gtk_widget_show (GTK_WIDGET (item));
-	gtk_tooltips_set_tip (tooltips, GTK_WIDGET (item),
-			      _("Shut down your computer so that "
-				"you may turn it off."),
-			      NULL);
+	if (bin_exists (GdmHalt)) {
+		item = gtk_menu_item_new_with_label (_("Halt..."));
+		gtk_menu_append (GTK_MENU (menu), item);
+		gtk_signal_connect (GTK_OBJECT (item), "activate",
+				    GTK_SIGNAL_FUNC (gdm_login_halt_handler), 
+				    NULL);
+		gtk_widget_show (GTK_WIDGET (item));
+		gtk_tooltips_set_tip (tooltips, GTK_WIDGET (item),
+				      _("Shut down your computer so that "
+					"you may turn it off."),
+				      NULL);
+		got_anything = TRUE;
+	}
+
+	if (bin_exists (GdmSuspend)) {
+		item = gtk_menu_item_new_with_label (_("Suspend..."));
+		gtk_menu_append (GTK_MENU (menu), item);
+		gtk_signal_connect (GTK_OBJECT (item), "activate",
+				    GTK_SIGNAL_FUNC (gdm_login_suspend_handler), 
+				    NULL);
+		gtk_widget_show (GTK_WIDGET (item));
+		gtk_tooltips_set_tip (tooltips, GTK_WIDGET (item),
+				      _("Suspend your computer"),
+				      NULL);
+		got_anything = TRUE;
+	}
 	
-	item = gtk_menu_item_new_with_label (_("System"));
-	gtk_menu_bar_append (GTK_MENU_BAR (menubar), item);
-	gtk_menu_item_set_submenu (GTK_MENU_ITEM (item), menu);
-        gtk_widget_add_accelerator (item, "activate_item", accel,
-				    GDK_y, GDK_MOD1_MASK, 0);
-	gtk_widget_show (GTK_WIDGET (item));
+	if (got_anything) {
+		item = gtk_menu_item_new_with_label (_("System"));
+		gtk_menu_bar_append (GTK_MENU_BAR (menubar), item);
+		gtk_menu_item_set_submenu (GTK_MENU_ITEM (item), menu);
+		gtk_widget_add_accelerator (item, "activate_item", accel,
+					    GDK_y, GDK_MOD1_MASK, 0);
+		gtk_widget_show (GTK_WIDGET (item));
+	}
+    }
+
+    /* Add a quit item when in xdmcp mode */
+    if (gdm_string_empty (g_getenv ("GDM_IS_LOCAL"))) {
+	    item = gtk_menu_item_new_with_label (_("Quit"));
+	    gtk_menu_bar_append (GTK_MENU_BAR (menubar), item);
+	    gtk_widget_add_accelerator (item, "activate_item", accel,
+					GDK_q, GDK_MOD1_MASK, 0);
+	    gtk_widget_show (GTK_WIDGET (item));
+	    gtk_signal_connect (GTK_OBJECT (item), "activate",
+				GTK_SIGNAL_FUNC (gtk_main_quit), NULL);
     }
 
     /* The clock */
@@ -2638,13 +2766,13 @@ gdm_login_gui_init (void)
 	/* FIXME: do smarter sizing here */
 	width = maxwidth + (maxwidth + 20) * (number_of_users < 5 ?
 					      number_of_users : 5);
-	if (width > screen.width * 0.5)
-		width = screen.width * 0.5;
+	if (width > gdm_wm_screen.width * 0.5)
+		width = gdm_wm_screen.width * 0.5;
 
 	height = (maxheight * 1.9 ) *
 		(1 + (number_of_users - 1) / 5);
-	if (height > screen.height * 0.25)
-		height = screen.height * 0.25;
+	if (height > gdm_wm_screen.height * 0.25)
+		height = gdm_wm_screen.height * 0.25;
 
 	gtk_widget_set_usize (GTK_WIDGET (bbox), width, height);
     }
@@ -2677,12 +2805,12 @@ gdm_login_gui_init (void)
 
 		gdk_window_get_size ((GdkWindow *) GNOME_PIXMAP (logo)->pixmap,
 				     &lw, &lh);
-		if (lw > screen.width / 2)
-			lw = screen.width / 2;
+		if (lw > gdm_wm_screen.width / 2)
+			lw = gdm_wm_screen.width / 2;
 		else
 			lw = -1;
-		if (lh > (2 * screen.height) / 3)
-			lh = (2 * screen.height) / 3;
+		if (lh > (2 * gdm_wm_screen.height) / 3)
+			lh = (2 * gdm_wm_screen.height) / 3;
 		else
 			lh = -1;
 		if (lw > -1 || lh > -1)
@@ -2808,7 +2936,7 @@ gdm_login_gui_init (void)
 				GTK_SIGNAL_FUNC (login_realized),
 				NULL);
     } else {
-	    gdm_center_window (GTK_WINDOW (login));
+	    gdm_wm_center_window (GTK_WINDOW (login));
     }
 
     gtk_widget_show_all (GTK_WIDGET (login));
@@ -2930,6 +3058,7 @@ gdm_login_check_exclude (struct passwd *pwent)
 		excludes = g_strsplit (GdmExclude, ",", 0);
 
 		for (i=0 ; excludes[i] != NULL ; i++)  {
+			g_strstrip (excludes[i]);
 			if (strcasecmp_no_locale (excludes[i],
 						  pwent->pw_name) == 0) {
 				g_strfreev (excludes);
@@ -3005,89 +3134,6 @@ gdm_login_users_init (void)
     endpwent ();
 }
 
-
-static void 
-gdm_screen_init (void) 
-{
-#ifdef HAVE_LIBXINERAMA
-	gboolean have_xinerama = FALSE;
-
-	gdk_flush ();
-	gdk_error_trap_push ();
-	have_xinerama = XineramaIsActive (GDK_DISPLAY ());
-	gdk_flush ();
-	if (gdk_error_trap_pop () != 0)
-		have_xinerama = FALSE;
-
-	if (have_xinerama) {
-		int screen_num, i;
-		XineramaScreenInfo *xscreens =
-			XineramaQueryScreens (GDK_DISPLAY (),
-					      &screen_num);
-
-
-		if (screen_num <= 0) {
-			/* should NEVER EVER happen */
-			syslog (LOG_ERR, "Xinerama active, but <= 0 screens?");
-			screen.x = 0;
-			screen.y = 0;
-			screen.width = gdk_screen_width ();
-			screen.height = gdk_screen_height ();
-
-			allscreens = g_new0 (GdkRectangle, 1);
-			allscreens[0] = screen;
-			screens = 1;
-			return;
-		}
-
-		if (screen_num <= GdmXineramaScreen)
-			GdmXineramaScreen = 0;
-
-		allscreens = g_new0 (GdkRectangle, screen_num);
-		screens = screen_num;
-
-		for (i = 0; i < screen_num; i++) {
-			allscreens[i].x = xscreens[i].x_org;
-			allscreens[i].y = xscreens[i].y_org;
-			allscreens[i].width = xscreens[i].width;
-			allscreens[i].height = xscreens[i].height;
-
-			if (GdmXineramaScreen == i)
-				screen = allscreens[i];
-		}
-
-		XFree (xscreens);
-	} else
-#endif
-	{
-		screen.x = 0;
-		screen.y = 0;
-		screen.width = gdk_screen_width ();
-		screen.height = gdk_screen_height ();
-
-		allscreens = g_new0 (GdkRectangle, 1);
-		allscreens[0] = screen;
-		screens = 1;
-	}
-#if 0
-	/* for testing Xinerama support on non-xinerama setups */
-	{
-		screen.x = 100;
-		screen.y = 100;
-		screen.width = gdk_screen_width () / 2 - 100;
-		screen.height = gdk_screen_height () / 2 - 100;
-
-		allscreens = g_new0 (GdkRectangle, 2);
-		allscreens[0] = screen;
-		allscreens[1].x = gdk_screen_width () / 2;
-		allscreens[1].y = gdk_screen_height () / 2;
-		allscreens[1].width = gdk_screen_width () / 2;
-		allscreens[1].height = gdk_screen_height () / 2;
-		screens = 2;
-	}
-#endif
-}
-
 static void
 set_root (GdkPixbuf *pb)
 {
@@ -3133,16 +3179,16 @@ render_scaled_back (const GdkPixbuf *pb)
 	width = gdk_pixbuf_get_width (pb);
 	height = gdk_pixbuf_get_height (pb);
 
-	for (i = 0; i < screens; i++) {
+	for (i = 0; i < gdm_wm_screens; i++) {
 		gdk_pixbuf_scale (pb, back,
-				  allscreens[i].x,
-				  allscreens[i].y,
-				  allscreens[i].width,
-				  allscreens[i].height,
-				  allscreens[i].x /* offset_x */,
-				  allscreens[i].y /* offset_y */,
-				  (double) allscreens[i].width / width,
-				  (double) allscreens[i].height / height,
+				  gdm_wm_allscreens[i].x,
+				  gdm_wm_allscreens[i].y,
+				  gdm_wm_allscreens[i].width,
+				  gdm_wm_allscreens[i].height,
+				  gdm_wm_allscreens[i].x /* offset_x */,
+				  gdm_wm_allscreens[i].y /* offset_y */,
+				  (double) gdm_wm_allscreens[i].width / width,
+				  (double) gdm_wm_allscreens[i].height / height,
 				  GDK_INTERP_BILINEAR);
 	}
 
@@ -3256,6 +3302,7 @@ main (int argc, char *argv[])
     struct sigaction chld;
     sigset_t mask;
     GIOChannel *ctrlch;
+    char *gdm_version;
 
     if (g_getenv ("DOING_GDM_DEVELOPMENT") != NULL)
 	    DOING_GDM_DEVELOPMENT = TRUE;
@@ -3293,7 +3340,113 @@ main (int argc, char *argv[])
 
     tooltips = gtk_tooltips_new ();
 
-    gdm_screen_init ();
+    gdm_wm_screen_init (GdmXineramaScreen);
+
+    gdm_version = g_getenv ("GDM_VERSION");
+
+    if ( ! DOING_GDM_DEVELOPMENT &&
+	(gdm_version == NULL ||
+	 strcmp (gdm_version, VERSION) != 0) &&
+	gdm_string_empty (g_getenv ("GDM_IS_LOCAL"))) {
+	    char *msg;
+	    GtkWidget *dialog;
+
+	    gdm_wm_init (0);
+
+	    gdm_wm_focus_new_windows (TRUE);
+
+	    msg = g_strdup_printf
+		    (_("The greeter version (%s) does not match the daemon "
+		       "version.\n"
+		       "You have probably just upgraded gdm.\n"
+		       "Please restart the gdm daemon or reboot the computer."),
+		     VERSION);
+	    dialog = gnome_error_dialog (msg);
+	    g_free (msg);
+
+	    gtk_widget_show_all (dialog);
+	    gdm_wm_center_window (GTK_WINDOW (dialog));
+	    gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
+
+	    gnome_dialog_run_and_close (GNOME_DIALOG (dialog));
+
+	    return EXIT_SUCCESS;
+    }
+
+    if ( ! DOING_GDM_DEVELOPMENT &&
+	gdm_version == NULL) {
+	    char *msg;
+	    GtkWidget *dialog;
+
+	    gdm_wm_init (0);
+
+	    gdm_wm_focus_new_windows (TRUE);
+
+	    msg = g_strdup_printf
+		    (_("The greeter version (%s) does not match the daemon "
+		       "version.\n"
+		       "You have probably just upgraded gdm.\n"
+		       "Please restart the gdm daemon or reboot the computer."),
+		     VERSION);
+	    dialog = gnome_message_box_new (msg,
+					    GNOME_MESSAGE_BOX_WARNING,
+					    _("Reboot"),
+					    GNOME_STOCK_BUTTON_CLOSE,
+					    NULL);
+	    g_free (msg);
+
+	    gtk_widget_show_all (dialog);
+	    gdm_wm_center_window (GTK_WINDOW (dialog));
+	    gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
+
+	    switch (gnome_dialog_run_and_close (GNOME_DIALOG (dialog))) {
+	    case 0:
+		    return DISPLAY_REBOOT;
+	    default:
+		    return DISPLAY_ABORT;
+	    }
+    }
+
+    if ( ! DOING_GDM_DEVELOPMENT &&
+	strcmp (gdm_version, VERSION) != 0) {
+	    char *msg;
+	    GtkWidget *dialog;
+
+	    gdm_wm_init (0);
+
+	    gdm_wm_focus_new_windows (TRUE);
+
+	    msg = g_strdup_printf
+		    (_("The greeter version (%s) does not match the daemon "
+		       "version (%s).\n"
+		       "You have probably just upgraded gdm.\n"
+		       "Please restart the gdm daemon or reboot the computer."),
+		     VERSION, gdm_version);
+	    dialog = gnome_message_box_new (msg,
+					    GNOME_MESSAGE_BOX_WARNING,
+					    _("Restart"),
+					    _("Reboot"),
+					    GNOME_STOCK_BUTTON_CLOSE,
+					    NULL);
+	    g_free (msg);
+
+	    gtk_widget_show_all (dialog);
+	    gdm_wm_center_window (GTK_WINDOW (dialog));
+	    gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
+	    gtk_widget_show_now (dialog);
+
+	    gnome_dialog_set_default (GNOME_DIALOG (dialog), 0);
+	    gnome_dialog_grab_focus (GNOME_DIALOG (dialog), 0);
+
+	    switch (gnome_dialog_run_and_close (GNOME_DIALOG (dialog))) {
+	    case 0:
+		    return DISPLAY_RESTARTGDM;
+	    case 1:
+		    return DISPLAY_REBOOT;
+	    default:
+		    return DISPLAY_ABORT;
+	    }
+    }
 
     if (GdmBrowser)
 	gdm_login_users_init ();
@@ -3392,7 +3545,7 @@ main (int argc, char *argv[])
 		       "There are two available sessions you can use, but\n"
 		       "you should log in and correct the gdm configuration."));
 	    gtk_widget_show_all (dialog);
-	    gdm_center_window (GTK_WINDOW (dialog));
+	    gdm_wm_center_window (GTK_WINDOW (dialog));
 	    gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
 
 	    gnome_dialog_run_and_close (GNOME_DIALOG (dialog));
@@ -3408,7 +3561,7 @@ main (int argc, char *argv[])
 		       "line for the login dialog, and thus I ran the\n"
 		       "default command.  Please fix your configuration."));
 	    gtk_widget_show_all (dialog);
-	    gdm_center_window (GTK_WINDOW (dialog));
+	    gdm_wm_center_window (GTK_WINDOW (dialog));
 	    gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
 
 	    gnome_dialog_run_and_close (GNOME_DIALOG (dialog));
@@ -3426,12 +3579,13 @@ main (int argc, char *argv[])
 		       "and create a configuration file with the GDM\n"
 		       "configuration program."));
 	    gtk_widget_show_all (dialog);
-	    gdm_center_window (GTK_WINDOW (dialog));
+	    gdm_wm_center_window (GTK_WINDOW (dialog));
 	    gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
 
 	    gnome_dialog_run_and_close (GNOME_DIALOG (dialog));
     }
 
+    g_atexit (kill_thingies);
 
     gtk_main ();
 
